/*	Copyright (C) 2003-2013 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#include "maya/maya.pmh"

#define FE_JCO_DEBUG		FALSE

using namespace fe;

void JointCombineOp::initialize(void)
{
	catalog<String>("icon")="FE_alpha";

	catalog<String>("root")="joints";
	catalog<String>("root","label")="Root Node Name";

	//* not read; just for update dependency
	catalog< sp<Component> >("Input Trigger");

	//* no copy
	catalog< sp<Component> >("Output Surface");
	catalog<String>("Output Surface","IO")="output";

	m_spScope=registry()->create("Scope");
}

void JointCombineOp::handle(Record& a_rSignal)
{
	const String rootName=catalog<String>("root");

#if FE_JCO_DEBUG
	feLog("JointCombineOp::handle\n");
#endif

	sp<SurfaceAccessibleI> spOutputAccessible;
	if(!accessOutput(spOutputAccessible,a_rSignal)) return;

	m_spJointAccessible=registry()->create("*.SurfaceAccessibleJoint");
	if(m_spJointAccessible.isNull())
	{
		return;
	}

	m_spJointAccessible->bind(m_spScope);
	m_spJointAccessible->startJoints();

	//* TODO is there an easier way to get a node object by name?
	MSelectionList originalList;
	MGlobal::getActiveSelectionList(originalList);

	MGlobal::selectByName(rootName.c_str(),MGlobal::kReplaceList);

	MSelectionList selectionList;
	MGlobal::getActiveSelectionList(selectionList);

	MGlobal::setActiveSelectionList(originalList);

	MStatus status;
	MObject rootObject;
	if(selectionList.length())
	{
		selectionList.getDependNode(0,rootObject);
	}
	if(rootObject.isNull())
	{
		String message;
		message.sPrintf("root node not found: %s",
				status.errorString().asChar());
		catalog<String>("error")+=message;
		return;
	}

	m_animWorldMap.clear();
	setIdentity(m_animWorldMap[""]);

	m_refWorldMap.clear();
	setIdentity(m_refWorldMap[""]);

	scanDag(rootObject,"");

	spOutputAccessible->copy(m_spJointAccessible);

#if FE_JCO_DEBUG
	feLog("JointCombineOp::handle done\n");
#endif
}

void JointCombineOp::scanDag(MObject a_parent,String a_grandparent)
{
	const Color yellow(1.0,1.0,0.0);

	if(a_parent.isNull())
	{
		return;
	}

	MFnDagNode dagParent(a_parent);
	const String name=dagParent.name().asChar();

	SpatialTransform xformLocal;
	SpatialTransform xformLocalRef;

	MStatus status;
	MFnIkJoint fnJoint(a_parent,&status);
	if(status)
	{
		//* matrix = [S] * [RO] * [R] * [JO] * [IS] * [T]

		//* [S]  getScale()
		//* [RO] getScaleOrientation()
		//* [R]  getRotation()
		//* [JO] getOrientation()
		//* [IS] inverse of the getScale on the parent transformation matrix
		//* [T]  getTranslation()

		//* TODO scale

		const MVector trans=fnJoint.getTranslation(MSpace::kTransform);

		MQuaternion scaleOrientation;
		status=fnJoint.getScaleOrientation(scaleOrientation);

		MQuaternion rotation;
		status=fnJoint.getRotation(rotation,MSpace::kTransform);

		MQuaternion orientation;
		status=fnJoint.getOrientation(orientation);

		const SpatialVector relativeTranslation(trans[0],trans[1],trans[2]);
		const SpatialQuaternion relativeScaleOrientation(
				scaleOrientation[0],scaleOrientation[1],
				scaleOrientation[2],scaleOrientation[3]);
		const SpatialQuaternion relativeRotation(
				rotation[0],rotation[1],rotation[2],rotation[3]);
		const SpatialQuaternion relativeOrientation(
				orientation[0],orientation[1],
				orientation[2],orientation[3]);

		SpatialTransform relativeMove;
		setIdentity(relativeMove);
		translate(relativeMove,relativeTranslation);

		xformLocalRef=
				SpatialTransform(relativeScaleOrientation)*
				SpatialTransform(relativeOrientation)*
				relativeMove;

		xformLocal=
				SpatialTransform(relativeScaleOrientation)*
				SpatialTransform(relativeRotation)*
				SpatialTransform(relativeOrientation)*
				relativeMove;

#if FE_JCO_DEBUG
		if(name=="l_uparm")
		{
			const SpatialEuler euler=SpatialTransform(relativeRotation);

			feLog("for \"%s\" t %s r %s e %s\n",
					dagParent.name().asChar(),
					c_print(relativeTranslation),
					c_print(relativeRotation));
			feLog("  e %.6G %.6G %.6G ro %s o %s\n%s\n",
					euler[0]/degToRad,
					euler[1]/degToRad,
					euler[2]/degToRad,
					c_print(relativeScaleOrientation),
					c_print(relativeOrientation),
					c_print(xformLocal));
		}
#endif
	}
	else
	{
		setIdentity(xformLocal);
	}

	const SpatialTransform animWorld=xformLocal*m_animWorldMap[a_grandparent];

	m_animWorldMap[name]=animWorld;

	//* TODO ref is orientation only

	const SpatialTransform refWorld=xformLocalRef*m_refWorldMap[a_grandparent];

	m_refWorldMap[name]=refWorld;

	if(!a_grandparent.empty())
	{
		m_spJointAccessible->addJoint(name,a_grandparent,
				xformLocal,animWorld,refWorld,yellow);
	}

	const U32 childCount=dagParent.childCount(&status);
	for(U32 childIndex=0;childIndex<childCount;childIndex++)
	{
		MObject child=dagParent.child(childIndex,&status);
		MFnDagNode dagChild(child);
//		feLog("parent \"%s\" child \"%s\"\n",
//				dagParent.name().asChar(),dagChild.name().asChar());

		scanDag(child,name);
	}
}
