/*	Copyright (C) 2003-2013 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#include "maya/maya.pmh"

#define FE_JSO_DEBUG		FALSE

using namespace fe;

void JointSeparateOp::initialize(void)
{
	catalog<String>("icon")="FE_alpha";

	catalog<bool>("active")=true;
	catalog<String>("active","label")="Update Joints";

	catalog<bool>("useRef")=false;
	catalog<String>("useRef","label")="Use Reference";

	catalog<String>("root")="joints";
	catalog<String>("root","label")="Root Node Name";

	catalog< sp<Component> >("Input Surface");

	//* hook up root joint geometry attribute
	catalog< sp<Component> >("Output Surface");
	catalog<String>("Output Surface","IO")="output";
//	catalog<String>("Output Surface","implementation")="curve";
	catalog<bool>("Output Surface","array")=true;
}

void JointSeparateOp::handle(Record& a_rSignal)
{
	if(!catalog<bool>("active"))
	{
		return;
	}

	const String rootName=catalog<String>("root");
	const BWORD useRef=catalog<bool>("useRef");

#if FE_JSO_DEBUG
	feLog("JointSeparateOp::handle\n");
#endif

	sp<SurfaceAccessibleI> spInputAccessible;
	if(!access(spInputAccessible,"Input Surface")) return;

	sp<SurfaceAccessorI> spInputName;
	if(!access(spInputName,spInputAccessible,e_primitive,"name")) return;

	sp<SurfaceAccessorI> spInputParent;
	if(!access(spInputParent,spInputAccessible,e_primitive,"parent")) return;

	sp<SurfaceAccessorI> spInputAnimX;
	if(!access(spInputAnimX,spInputAccessible,e_primitive,"animX")) return;

	sp<SurfaceAccessorI> spInputAnimY;
	if(!access(spInputAnimY,spInputAccessible,e_primitive,"animY")) return;

	sp<SurfaceAccessorI> spInputAnimZ;
	if(!access(spInputAnimZ,spInputAccessible,e_primitive,"animZ")) return;

	sp<SurfaceAccessorI> spInputAnimT;
	if(!access(spInputAnimT,spInputAccessible,e_primitive,"animT")) return;

	sp<SurfaceAccessorI> spInputRefX;
	if(!access(spInputRefX,spInputAccessible,e_primitive,"refX")) return;

	sp<SurfaceAccessorI> spInputRefY;
	if(!access(spInputRefY,spInputAccessible,e_primitive,"refY")) return;

	sp<SurfaceAccessorI> spInputRefZ;
	if(!access(spInputRefZ,spInputAccessible,e_primitive,"refZ")) return;

	sp<SurfaceAccessorI> spInputRefT;
	if(!access(spInputRefT,spInputAccessible,e_primitive,"refT")) return;

	//* TODO is there an easier way to get a node object by name?
	MSelectionList originalList;
	MGlobal::getActiveSelectionList(originalList);

	MGlobal::selectByName(rootName.c_str(),MGlobal::kReplaceList);

	MSelectionList selectionList;
	MGlobal::getActiveSelectionList(selectionList);

	MGlobal::setActiveSelectionList(originalList);

	MStatus status;
	MObject rootObject;
	if(selectionList.length())
	{
		selectionList.getDependNode(0,rootObject);
	}
	if(rootObject.isNull())
	{
		MFnTransform fnTransform;
		rootObject=fnTransform.create(MObject::kNullObj,&status);
		if(!status)
		{
			feLog("JointSeparateOp::handle create root failed: %s\n",
					status.errorString().asChar());
			return;
		}
		fnTransform.setName(rootName.c_str());
	}

	const U32 primitiveCount=spInputName->count();

#if FE_JSO_DEBUG
	feLog("JointSeparateOp::handle primitiveCount %d\n",primitiveCount);
#endif

	std::map<String,MObject> objectMap;

	m_keepList.clear();

	for(U32 primitiveIndex=0;primitiveIndex<primitiveCount;primitiveIndex++)
	{
#if FE_JSO_DEBUG
		feLog("JointSeparateOp::handle %d/%d\n",primitiveIndex,primitiveCount);
#endif

		const String name=spInputName->string(primitiveIndex);

		const String parentName=spInputParent->string(primitiveIndex);
		MObject parentObject=objectMap[parentName];
		if(parentObject==MObject::kNullObj)
		{
			parentObject=rootObject;
		}

		SpatialTransform xformLocalRef;
		xformLocalRef.column(0)=spInputRefX->spatialVector(primitiveIndex);
		xformLocalRef.column(1)=spInputRefY->spatialVector(primitiveIndex);
		xformLocalRef.column(2)=spInputRefZ->spatialVector(primitiveIndex);
		xformLocalRef.column(3)=spInputRefT->spatialVector(primitiveIndex);

		SpatialTransform xformLocal;
		if(useRef)
		{
			xformLocal=xformLocalRef;
		}
		else
		{
			xformLocal.column(0)=spInputAnimX->spatialVector(primitiveIndex);
			xformLocal.column(1)=spInputAnimY->spatialVector(primitiveIndex);
			xformLocal.column(2)=spInputAnimZ->spatialVector(primitiveIndex);
			xformLocal.column(3)=spInputAnimT->spatialVector(primitiveIndex);
		}

		SpatialTransform invLocalRef;
		invert(invLocalRef,xformLocalRef);

		const SpatialVector localDisplace=xformLocal.translation();

		const SpatialQuaternion quatRotation=xformLocal*invLocalRef;
		const SpatialQuaternion quatOrientation=xformLocalRef;

		const MVector mDisplace(
				localDisplace[0],localDisplace[1],localDisplace[2]);
		const MQuaternion mRotation(
				quatRotation[0],quatRotation[1],
				quatRotation[2],quatRotation[3]);
		const MQuaternion mOrientation(
				quatOrientation[0],quatOrientation[1],
				quatOrientation[2],quatOrientation[3]);

#if FE_JSO_DEBUG
		feLog("  \"%s\" parent \"%s\"\n",name.c_str(),parentName.c_str());
		feLog("  point %s localDisplace %s\n",
				c_print(xformLocal.translation()),c_print(localDisplace));
		feLog("  quatRotation %s\n",c_print(quatRotation));
		feLog("  quatOrientation %s\n",c_print(quatOrientation));
#endif

		MObject jointObject;

		MFnDagNode fnDagNode(parentObject);
		const U32 childCount=fnDagNode.childCount(&status);
		for(U32 childIndex=0;childIndex<childCount;childIndex++)
		{
			MObject child=fnDagNode.child(childIndex,&status);
			MFnDagNode dagChild(child);
			if(dagChild.name()==name.c_str())
			{
				jointObject=child;
				break;
			}
		}

		MFnIkJoint fnJoint(jointObject);

		if(jointObject.isNull())
		{
			jointObject=fnJoint.create(parentObject,&status);
			if(!status)
			{
				feLog("JointSeparateOp::handle create failed: %s\n",
						status.errorString().asChar());
				continue;
			}

			fnJoint.setName(name.c_str());
		}

		objectMap[name]=jointObject;
		m_keepList.push_back(jointObject);

		fnJoint.setTranslation(mDisplace,MSpace::kTransform);
		fnJoint.setRotation(mRotation,MSpace::kTransform);
		fnJoint.setOrientation(mOrientation);
	}

	//* delete unused pre-existing joints
	scanDag(rootObject);

#if FE_JSO_DEBUG
	feLog("JointSeparateOp::handle done\n");
#endif
}

void JointSeparateOp::scanDag(MObject a_parent)
{
	if(a_parent.isNull())
	{
		return;
	}

	MFnDagNode dagParent(a_parent);

	std::vector<MObject> deleteList;

	MStatus status;
	const U32 childCount=dagParent.childCount(&status);
	for(U32 childIndex=0;childIndex<childCount;childIndex++)
	{
		MObject child=dagParent.child(childIndex,&status);

		const U32 keepCount=m_keepList.size();
		U32 keepIndex=0;
		for(;keepIndex<keepCount;keepIndex++)
		{
			if(m_keepList[keepIndex]==child)
			{
				scanDag(child);
				break;
			}
		}
		if(keepIndex==keepCount)
		{
			deleteList.push_back(child);
		}
	}

	const U32 deleteCount=deleteList.size();
	for(U32 deleteIndex=0;deleteIndex<deleteCount;deleteIndex++)
	{
		MObject child=deleteList[deleteIndex];

		MGlobal::deleteNode(child);
	}
}
