/*	Copyright (C) 2003-2013 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

//* NOTE refinement with multithreading will crash

#define FE_LCO_DEBUG	FALSE

#include "operator/operator.pmh"

using namespace fe;

//* TODO Extrapolating pre-scale - rescale whole curve along
//* root to tip vector to restore curve length

void LengthCorrectOp::initialize(void)
{
	catalog<I32>("Iterations")=1;
	catalog<I32>("Iterations","high")=10;
	catalog<I32>("Iterations","max")=100;

	catalog<bool>("Root Lock")=false;
	catalog<String>("Root Lock","hint")=
			"Leave first vertex of each primitive untouched."
			"  This can lock down rooted curves, like hair.";

	catalog<Real>("Tension Rate")=1.0;
	catalog<Real>("Tension Rate","high")=1.0;
	catalog<String>("Tension Rate","hint")=
			"Influence how quickly each iteration pulls towards"
			" the correct length.";

	catalog<Real>("Compression Rate")=1.0;
	catalog<Real>("Compression Rate","high")=1.0;
	catalog<String>("Compression Rate","hint")=
			"Influence how quickly each iteration pushes towards"
			" the correct length.";

	catalog<Real>("Correction Bias")=0.5;
	catalog<Real>("Correction Bias","max")=1.0;
	catalog<String>("Correction Bias","hint")=
			"Influence how much each point on an edge gets changed."
			"  Try lower numbers for rooted curves, like 0.3,"
			" but generally leave at 0.5 for polygons.";

	catalog<Real>("Suspension")=0.0;
	catalog<String>("Suspension","label")="Sliding Snap";
	catalog<Real>("Suspension","max")=1.0;
	catalog<String>("Suspension","hint")=
			"Restrain changes to slide along original input.";

	catalog<bool>("Collide")=false;
	catalog<String>("Collide","label")="Collision Avoidance";
	catalog<bool>("Collide","joined")=true;
	catalog<String>("Collide","hint")=
			"Push resulting surface out of collider.";

	catalog<Real>("Margin")=0.1;
	catalog<bool>("Margin","joined")=true;
	catalog<Real>("Margin","high")=1.0;
	catalog<Real>("Margin","max")=1e3;
	catalog<String>("Margin","enabler")="Collide";
	catalog<String>("Margin","hint")=
			"Minimum distance to keep from collider surface.";

	catalog<Real>("CollideRate")=1.0;
	catalog<String>("CollideRate","label")="Rate";
	catalog<Real>("CollideRate","high")=1.0;
	catalog<String>("CollideRate","enabler")="Collide";
	catalog<String>("CollideRate","hint")=
			"Amount of collision response per iteration.";

	catalog<I32>("Refinement")=0;
	catalog<String>("Refinement","label")="Collider Refinement";
	catalog<String>("Refinement","enabler")="Collide";

	catalog<bool>("Locking")=false;
	catalog<bool>("Locking","joined")=true;
	catalog<String>("Locking","hint")=
			"Use locking attribute to resist change.";

	catalog<String>("LockAttr")="lock";
	catalog<String>("LockAttr","label")="Lock Attribute";
	catalog<String>("LockAttr","enabler")="Locking";
	catalog<String>("LockAttr","hint")=
			"Name of real attribute describing resistance to change"
			" (1.0 is full lock).";

	catalog< sp<Component> >("Input Surface");

	catalog< sp<Component> >("Reference Surface");

	catalog< sp<Component> >("Collider Surface");
	catalog<bool>("Collider Surface","optional")=true;
}

void LengthCorrectOp::handle(Record& a_rSignal)
{
#if FE_LCO_DEBUG
	feLog("LengthCorrectOp::handle\n");
#endif

	const Real suspension=catalog<Real>("Suspension");
	const BWORD collide=catalog<bool>("Collide");

	//* output surface is copy of input surface
	if(!accessOutput(m_spOutputAccessible,a_rSignal)) return;

	if(!access(m_spInputAccessible,"Input Surface")) return;

	if(!access(m_spReferenceAccessible,"Reference Surface",e_quiet))
	{
		m_spReferenceAccessible=m_spInputAccessible;

		catalog<String>("warning")=
				"Substituting Input Surface for missing Reference Surface.";
	}

	sp<SurfaceAccessorI> spOutputVertices;
	if(!access(spOutputVertices,m_spOutputAccessible,
			e_primitive,e_vertices)) return;

	sp<SurfaceAccessorI> spRefVertices;
	if(!access(spRefVertices,m_spReferenceAccessible,
			e_primitive,e_vertices)) return;

	if(suspension>0.0)
	{
		//* access input as driver surface
		if(!access(m_spDriver,m_spInputAccessible)) return;

		m_spDriver->setRefinement(0);
		m_spDriver->prepareForSearch();
	}
	else
	{
		m_spDriver=NULL;
	}

	m_spCollider=NULL;
	if(collide)
	{
		access(m_spCollider,"Collider Surface",e_warning);
		if(m_spCollider.isValid())
		{
			m_spCollider->setRefinement(catalog<I32>("Refinement"));
			m_spCollider->prepareForSearch();
		}
	}

	const U32 primitiveCount=spOutputVertices->count();
	const U32 primitiveCountRef=spRefVertices->count();
	if(primitiveCount!=primitiveCountRef)
	{
		feLog("LengthCorrectOp::handle"
				" mismatched primitive count %d vs %d ref\n",
				primitiveCount,primitiveCountRef);
		catalog<String>("error")="mismatched reference surface";
		return;
	}

#if FE_LCO_DEBUG
	feLog("LengthCorrectOp::handle primitives %d\n", primitiveCount);
#endif

#if FALSE
	for(U32 iteration=0;iteration<iterations;iteration++)
	{
		//* tweak surface points towards reference length
		sp<SpannedRange> spFullRange=sp<SpannedRange>(new SpannedRange());
		spFullRange->addAtomic().add(0,primitiveCount);
		relength(-1,spFullRange);

		if(m_spDriver.isValid() || m_spCollider.isValid())
		{
			setAtomicRange(m_spOutputAccessible,e_pointsOnly);
			OperatorThreaded::handle(a_rSignal);
		}
	}
#else
	const I32 threadCount=catalog<I32>("Threads");
	if(threadCount>1)
	{
		sp<SpannedRange> spConnectivity=m_spOutputAccessible->atomize(
				SurfaceAccessibleI::e_pointsOfPrimitives,"",8*threadCount);

		setFullRange(spConnectivity);
		OperatorThreaded::handle(a_rSignal);
	}
	else
	{
		//* serial
		sp<SpannedRange> spFullRange=sp<SpannedRange>(new SpannedRange());
		spFullRange->nonAtomic().add(0,primitiveCount-1);
		run(-1,spFullRange);
	}
#endif

#if FE_LCO_DEBUG
	feLog("LengthCorrectOp::handle done\n");
#endif
}

void LengthCorrectOp::run(I32 a_id,sp<SpannedRange> a_spRange)
{
	const I32 iterations=catalog<I32>("Iterations");
	for(U32 iteration=0;iteration<iterations;iteration++)
	{
#if FE_LCO_DEBUG
		feLog("LengthCorrectOp::relength iteration %d/%d\n",
				iteration,iterations);
#endif

		relength(a_id,a_spRange);

		if(m_spDriver.isValid() || m_spCollider.isValid())
		{
			suspendAndCollide(a_id,a_spRange);
		}
	}
}

void LengthCorrectOp::relength(I32 a_id,sp<SpannedRange> a_spRange)
{
	const BWORD rootLock=catalog<bool>("Root Lock");
	const Real tensionRate=catalog<Real>("Tension Rate");
	const Real compressionRate=catalog<Real>("Compression Rate");
	const Real correctionBias=catalog<Real>("Correction Bias");
	const String lockAttr=catalog<String>("LockAttr");
	const BWORD locking=(catalog<bool>("Locking") && !lockAttr.empty());

	sp<SurfaceAccessorI> spInputProperties;
	if(!access(spInputProperties,m_spInputAccessible,
			e_primitive,e_properties)) return;

	sp<SurfaceAccessorI> spInputLock;
	if(locking)
	{
		access(spInputLock,m_spInputAccessible,e_point,lockAttr,e_warning);
	}

	sp<SurfaceAccessorI> spRefVertices;
	if(!access(spRefVertices,m_spReferenceAccessible,
			e_primitive,e_vertices)) return;

	sp<SurfaceAccessorI> spOutputVertices;
	if(!access(spOutputVertices,m_spOutputAccessible,
			e_primitive,e_vertices)) return;

	const U32 primitiveCount=spOutputVertices->count();

	for(SpannedRange::Iterator it=a_spRange->begin();!it.atEnd();it.step())
	{
		const U32 primitiveIndex=it.value();
		const BWORD openCurve=spInputProperties->integer(
				primitiveIndex,e_openCurve);
		const U32 subCount=spOutputVertices->subCount(primitiveIndex);
		const U32 subCountRef=spRefVertices->subCount(primitiveIndex);
		if(subCount!=subCountRef)
		{
			feLog("LengthCorrectOp::relength"
					" primitive %d mismatched subCount %d vs %d ref\n",
					primitiveIndex,subCount,subCountRef);
			continue;
		}

#if FE_LCO_DEBUG
		feLog("LengthCorrectOp::relength primitive %d/%d subCount %d\n",
				primitiveIndex,primitiveCount,subCount);
#endif

		//* NOTE skip root point with rootLock or any open curves
		for(U32 subIndex=(rootLock||openCurve);subIndex<subCount;subIndex++)
		{
			const U32 lastSubIndex=subIndex? subIndex-1: subCount-1;
			SpatialVector pointA=spOutputVertices->spatialVector(
					primitiveIndex,lastSubIndex);
			SpatialVector pointB=spOutputVertices->spatialVector(
					primitiveIndex,subIndex);

#if FE_LCO_DEBUG
			feLog("LengthCorrectOp::relength   %d/%d pointA %s pointB %s\n",
					subIndex,subCount,c_print(pointA),c_print(pointB));
#endif

			const SpatialVector delta=pointB-pointA;
			const Real length=magnitude(delta);
			if(length>0.0)
			{
				Real lockA=0.0;
				Real lockB=0.0;
				if(spInputLock.isValid())
				{
					const I32 vertexIndexA=spOutputVertices->integer(
							primitiveIndex,lastSubIndex);
					const I32 vertexIndexB=spOutputVertices->integer(
							primitiveIndex,subIndex);
					lockA=spInputLock->real(vertexIndexA);
					lockB=spInputLock->real(vertexIndexB);
				}
				const BWORD rootLocked=(rootLock && subIndex<2);
				if(rootLocked)
				{
					lockA=1.0;
				}

				const Real lockBias=0.5*(1.0+lockB-lockA);
				Real bias=lockBias+correctionBias-0.5;
				if(bias<0.0)
				{
					bias=0.0;
				}
				if(bias>1.0)
				{
					bias=1.0;
				}
				const Real antibias=1.0-bias;

				const SpatialVector pointRefA=spRefVertices->spatialVector(
						primitiveIndex,lastSubIndex);
				const SpatialVector pointRefB=
						spRefVertices->spatialVector(primitiveIndex,subIndex);
				const Real lengthRef=magnitude(pointRefB-pointRefA);

				const Real rate=(length>lengthRef)?
						tensionRate: compressionRate;
				const SpatialVector adjust=
						(lengthRef/length-1.0)*delta*rate;

#if FE_LCO_DEBUG
				feLog("LengthCorrectOp::relength   ref pointA %s pointB %s\n",
						c_print(pointRefA),c_print(pointRefB));
#endif

				if(rootLocked)
				{
					pointB+=(1.0-lockB)*adjust;
					spOutputVertices->set(primitiveIndex,subIndex,pointB);
				}
				else
				{
					pointA-=(1.0-lockA)*bias*adjust;
					pointB+=(1.0-lockB)*antibias*adjust;

					spOutputVertices->set(primitiveIndex,lastSubIndex,pointA);
					spOutputVertices->set(primitiveIndex,subIndex,pointB);
				}

#if FE_LCO_DEBUG
				feLog("LengthCorrectOp::relength"
						"   bias %.6G lock pointA %.6G pointB %.6G\n",
						bias,lockA,lockB);
				feLog("LengthCorrectOp::relength   now pointA %s pointB %s\n",
						c_print(pointA),c_print(pointB));
#endif
			}
		}
	}
}

void LengthCorrectOp::suspendAndCollide(I32 a_id,sp<SpannedRange> a_spRange)
{
	const Real suspension=catalog<Real>("Suspension");
	const String lockAttr=catalog<String>("LockAttr");
	const BWORD locking=(catalog<bool>("Locking") && !lockAttr.empty());
	const Real margin=catalog<Real>("Margin");
	const Real collideRate=catalog<Real>("CollideRate");

	sp<SurfaceAccessorI> spInputPoints;
	if(!access(spInputPoints,m_spInputAccessible,
			e_point,e_position)) return;

	sp<SurfaceAccessorI> spOutputPoints;
	if(!access(spOutputPoints,m_spOutputAccessible,
			e_point,e_position)) return;

	const U32 pointCount=spOutputPoints->count();
	BWORD pointChanged[pointCount];
	for(U32 pointIndex=0;pointIndex<pointCount;pointIndex++)
	{
		pointChanged[pointIndex]=FALSE;
	}

	sp<SurfaceAccessorI> spOutputVertices;
	if(!access(spOutputVertices,m_spOutputAccessible,
			e_primitive,e_vertices)) return;

	sp<SurfaceAccessorI> spInputLock;
	if(locking)
	{
		access(spInputLock,m_spInputAccessible,e_point,lockAttr,e_quiet);
	}

	SpatialVector driverCenter;
	Real driverRange2;
	if(m_spCollider.isValid())
	{
		driverCenter=m_spCollider->center();
		const Real range=m_spCollider->radius()+margin;
		driverRange2=range*range;
	}

	for(SpannedRange::Iterator it=a_spRange->begin();!it.atEnd();it.step())
	{
		const U32 primitiveIndex=it.value();

		const U32 subCount=spOutputVertices->subCount(primitiveIndex);
		for(U32 subIndex=0;subIndex<subCount;subIndex++)
		{
			const U32 pointIndex=
					spOutputVertices->integer(primitiveIndex,subIndex);
			BWORD& rChanged=pointChanged[pointIndex];
			if(rChanged)
			{
				continue;
			}

			SpatialVector point=spOutputPoints->spatialVector(pointIndex);
			Real lock=0.0;
			if(spInputLock.isValid())
			{
				lock=spInputLock->real(pointIndex);
				if(lock>=1.0)
				{
					continue;
				}
			}

			if(m_spDriver.isValid())
			{
				const SpatialVector original=
						spInputPoints->spatialVector(pointIndex);
				if(magnitudeSquared(point-original)>1e-6)
				{
					sp<SurfaceI::ImpactI> spImpact=
							m_spDriver->nearestPoint(point);
					if(spImpact.isValid())
					{
						point+=(1.0-lock)*suspension*
								(spImpact->intersection()-point);
						rChanged=TRUE;
					}
				}
			}

			if(m_spCollider.isValid())
			{
				//* TODO maxRadius arg to nearestPoint
				if(magnitudeSquared(point-driverCenter)<driverRange2)
				{
					sp<SurfaceI::ImpactI> spImpact=
							m_spCollider->nearestPoint(point);
					if(spImpact.isValid())
					{
						const SpatialVector norm=spImpact->normal();

						const SpatialVector target=
								spImpact->intersection()+margin*norm;

						const SpatialVector toTarget=target-point;
						const Real along=dot(toTarget,norm);
						if(along>0.0)
						{
#if FALSE
							feLog("LengthCorrectOp::run"
									"   point %s int %s\n norm %s along %.6G\n",
									c_print(point),
									c_print(spImpact->intersection()),
									c_print(norm),along);
#endif

							point+=collideRate*(1.0-lock)*along*norm;
							rChanged=TRUE;

#if FALSE
							feLog("   target %s to %s\n",
									c_print(toTarget),c_print(point));
#endif
						}
					}
				}
			}

			if(rChanged)
			{
				spOutputPoints->set(pointIndex,point);
			}

			//* do not repeat each point
			rChanged=TRUE;
		}
	}
}
